@@ -4,12 +4,20 @@ import inspect
 import logging
 from abc import ABC, abstractmethod
 from collections import OrderedDict
-from collections.abc import Callable, Iterable, Iterator, Mapping, MutableMapping
+from collections.abc import (
+    Callable,
+    Collection,
+    Iterable,
+    Iterator,
+    Mapping,
+    MutableMapping,
+)
 from contextlib import ContextDecorator, contextmanager, suppress
 from dataclasses import dataclass, field
 from enum import Enum, auto
 from functools import singledispatchmethod, wraps
 from inspect import Signature, get_annotations
+from threading import RLock
 from types import MappingProxyType
 from typing import (
     Any,
@@ -23,8 +31,8 @@ from typing import (
 )
 
 from injection.common.event import Event, EventChannel, EventListener
-from injection.common.formatting import format_type
 from injection.common.lazy import Lazy, LazyMapping
+from injection.common.tools import format_type, get_origin
 from injection.exceptions import (
     ModuleError,
     ModuleLockError,
@@ -35,6 +43,7 @@ from injection.exceptions import (
 __all__ = ("Injectable", "Module", "ModulePriorities")
 
 _logger = logging.getLogger(__name__)
+_thread_lock = RLock()
 
 _T = TypeVar("_T")
 
@@ -51,7 +60,7 @@ class ContainerEvent(Event, ABC):
 
 @dataclass(frozen=True, slots=True)
 class ContainerDependenciesUpdated(ContainerEvent):
-    classes: frozenset[type]
+    classes: Collection[type]
 
     def __str__(self) -> str:
         length = len(self.classes)
@@ -161,14 +170,16 @@ class SingletonInjectable(BaseInjectable[_T]):
         return self.__INSTANCE_KEY in self.cache
 
     def unlock(self):
-        self.cache.pop(self.__INSTANCE_KEY, None)
+        self.cache.clear()
 
     def get_instance(self) -> _T:
         with suppress(KeyError):
             return self.cache[self.__INSTANCE_KEY]
 
-        instance = self.factory()
-        self.cache[self.__INSTANCE_KEY] = instance
+        with _thread_lock:
+            instance = self.factory()
+            self.cache[self.__INSTANCE_KEY] = instance
+
         return instance
 
 
@@ -183,15 +194,15 @@ class Container:
     __channel: EventChannel = field(default_factory=EventChannel, init=False)
 
     def __getitem__(self, cls: type[_T], /) -> Injectable[_T]:
-        origin = self.__get_origin(cls)
+        origin = get_origin(cls)
 
         try:
             return self.__data[origin]
         except KeyError as exc:
             raise NoInjectable(cls) from exc
 
-    def __contains__(self, cls: type[_T], /) -> bool:
-        return self.__get_origin(cls) in self.__data
+    def __contains__(self, cls: type, /) -> bool:
+        return get_origin(cls) in self.__data
 
     @property
     def is_locked(self) -> bool:
@@ -201,8 +212,8 @@ class Container:
     def __injectables(self) -> frozenset[Injectable]:
         return frozenset(self.__data.values())
 
-    def set_multiple(self, classes: Iterable[type], injectable: Injectable):
-        classes = frozenset(self.__get_origin(cls) for cls in classes)
+    def update(self, classes: Iterable[type], injectable: Injectable):
+        classes = frozenset(get_origin(cls) for cls in classes)
 
         if classes:
             event = ContainerDependenciesUpdated(self, classes)
@@ -233,10 +244,6 @@ class Container:
     def notify(self, event: Event) -> ContextManager | ContextDecorator:
         return self.__channel.dispatch(event)
 
-    @staticmethod
-    def __get_origin(cls: type) -> type:
-        return getattr(cls, "__origin__", cls)
-
 
 """
 Module
@@ -280,11 +287,10 @@ class Module(EventListener):
 
         raise NoInjectable(cls)
 
-    def __setitem__(self, on: type | Iterable[type], injectable: Injectable, /):
-        classes = on if isinstance(on, Iterable) else (on,)
-        self.__container.set_multiple(classes, injectable)
+    def __setitem__(self, cls: type, injectable: Injectable, /):
+        self.update((cls,), injectable)
 
-    def __contains__(self, cls: type[_T], /) -> bool:
+    def __contains__(self, cls: type, /) -> bool:
         return any(cls in broker for broker in self.__brokers)
 
     def __str__(self) -> str:
@@ -328,7 +334,7 @@ class Module(EventListener):
 
     @property
     def __brokers(self) -> Iterator[Container | Module]:
-        yield from self.__modules
+        yield from tuple(self.__modules)
         yield self.__container
 
     def get_instance(self, cls: type[_T]) -> _T | None:
@@ -345,6 +351,10 @@ class Module(EventListener):
         instance = injectable.get_instance()
         return cast(cls, instance)
 
+    def update(self, classes: Iterable[type], injectable: Injectable):
+        self.__container.update(classes, injectable)
+        return self
+
     def use(
         self,
         module: Module,
@@ -614,7 +624,9 @@ class InjectableDecorator:
                 else:
                     yield on
 
-            self.__module[classes] = self.__injectable_type(wp)
+            injectable = self.__injectable_type(wp)
+            self.__module.update(classes, injectable)
+
             return wp
 
         return decorator(wrapped) if wrapped else decorator
